import random
import time
import requests

"""
requests 网络请求工具
"""
user_agents = ["Mozilla/5.0 (Windows NT 10.0; WOW64)", 'Mozilla/5.0 (Windows NT 6.3; WOW64)',
               'Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.11 (KHTML, like Gecko) Chrome/23.0.1271.64 Safari/537.11',
               'Mozilla/5.0 (Windows NT 6.3; WOW64; Trident/7.0; rv:11.0) like Gecko',
               'Mozilla/5.0 (Windows NT 5.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/28.0.1500.95 Safari/537.36',
               'Mozilla/5.0 (Windows NT 6.1; WOW64; Trident/7.0; SLCC2; .NET CLR 2.0.50727; .NET CLR 3.5.30729; .NET CLR 3.0.30729; Media Center PC 6.0; .NET4.0C; rv:11.0) like Gecko)',
               'Mozilla/5.0 (Windows; U; Windows NT 5.2) Gecko/2008070208 Firefox/3.0.1',
               'Mozilla/5.0 (Windows; U; Windows NT 5.1) Gecko/20070309 Firefox/2.0.0.3',
               'Mozilla/5.0 (Windows; U; Windows NT 5.1) Gecko/20070803 Firefox/1.5.0.12',
               'Mozilla/5.0 (Macintosh; PPC Mac OS X; U; en) Opera 8.0',
               'Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:1.8.1.12) Gecko/20080219 Firefox/2.0.0.12 Navigator/9.0.0.6',
               'Mozilla/4.0 (compatible; MSIE 8.0; Windows NT 6.1; Win64; x64; Trident/4.0)',
               'Mozilla/4.0 (compatible; MSIE 8.0; Windows NT 6.1; Trident/4.0)',
               'Mozilla/5.0 (compatible; MSIE 10.0; Windows NT 6.1; WOW64; Trident/6.0; SLCC2; .NET CLR 2.0.50727; .NET CLR 3.5.30729; .NET CLR 3.0.30729; Media Center PC 6.0; InfoPath.2; .NET4.0C; .NET4.0E)',
               'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.1 (KHTML, like Gecko) Maxthon/4.0.6.2000 Chrome/26.0.1410.43 Safari/537.1 ',
               'Mozilla/5.0 (compatible; MSIE 10.0; Windows NT 6.1; WOW64; Trident/6.0; SLCC2; .NET CLR 2.0.50727; .NET CLR 3.5.30729; .NET CLR 3.0.30729; Media Center PC 6.0; InfoPath.2; .NET4.0C; .NET4.0E; QQBrowser/7.3.9825.400)',
               'Mozilla/5.0 (Windows NT 6.1; WOW64; rv:21.0) Gecko/20100101 Firefox/21.0 ',
               'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.1 (KHTML, like Gecko) Chrome/21.0.1180.92 Safari/537.1 LBBROWSER',
               'Mozilla/5.0 (compatible; MSIE 10.0; Windows NT 6.1; WOW64; Trident/6.0; BIDUBrowser 2.x)',
               'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/536.11 (KHTML, like Gecko) Chrome/20.0.1132.11 TaoBrowser/3.0 Safari/536.11']


def get_html_text(url):
    request = None
    while True:
        try:
            # 在短時間使用多次requests会Failed to establish a new connection這樣一個錯誤，设置'Connection': 'close'
            headers = {'User-Agent': random.choice(user_agents), 'Connection': 'close'}
            request = requests.get(url, headers=headers, timeout=None)

            if request is not None:
                request.raise_for_status()
                request.encoding = request.apparent_encoding
                # print(r.url)
                if request.status_code == 200:
                    rst = request.text
                    request.close()
                    return rst
                request.close()
            return ""
        except Exception as e:
            request.close()
            print(f'请求发生错误：{e}')
            time.sleep(2)


def get_page_json(url):
    headers = {'cookie': '',
               'User-Agent': random.choice(user_agents)
               , 'Connection': 'close'}
    # url = 'https://www.csdn.net/api/articles?type=more&category=home&shown_offset=0'
    try:
        r = requests.get(url, headers=headers, timeout=30)
        r.raise_for_status()
        r.encoding = r.apparent_encoding
        if r.status_code == 200:
            return r.json()
        return ""
    except ConnectionError as e:
        print(e)
        return ""


def get_page_detail(url):
    headers = {
        'user-agent': random.choice(user_agents), 'Connection': 'close'}
    try:
        response = requests.get(url, headers=headers)
        if response.status_code == 200:
            return response.text
        return None
    except ConnectionError:
        print('Error occurred')
        return None


def download_image(url):
    print('Downloading', url)
    try:
        response = requests.get(url)
        if response.status_code == 200:
            save_image(response.content)
        return None
    except ConnectionError:
        return None


def save_image(content):
    import os
    file_path = '{0}'.format(os.getcwd() + '\images')
    if not os.path.exists(file_path):
        os.makedirs(file_path)
    from _md5 import md5
    image_path = '{0}/{1}.{2}'.format(os.getcwd() + '\images', md5(content).hexdigest(), 'jpg')
    if not os.path.exists(image_path):
        with open(image_path, 'wb') as f:
            f.write(content)
            f.close()



import logging
import os
from logging.handlers import TimedRotatingFileHandler
import logbook

logbook.set_datetime_format('local')


class DefaultLogHandler(object):
    """默认的 Log 类"""

    def __init__(self, filename='default.log', loglevel=logging.DEBUG):
        """Log对象
        :param :filename: log 文件名
        :param :loglevel: 设定log等级 ['CRITICAL', 'ERROR', 'WARNING', 'NOTICE', 'INFO', 'DEBUG', 'TRACE', 'NOTSET']
        """
        os.makedirs("./logs", exist_ok=True)
        self.log = self._logging(level=loglevel, filename="./logs/" + filename)

    def __getattr__(self, item, *args, **kwargs):
        return self.log.__getattribute__(item, *args, **kwargs)

    def _logging(self, **kwargs):
        level = kwargs.pop('level', None)
        filename = kwargs.pop('filename', None)
        datefmt = kwargs.pop('datefmt', None)
        format = kwargs.pop('format', None)
        if level is None:
            level = logging.DEBUG
        if filename is None:
            filename = 'default.log'
        if datefmt is None:
            datefmt = '%Y-%m-%d %H:%M:%S'
        if format is None:
            # [%(asctime)s] [%(levelname)s] [%(module)s] [%(lineno)d] %(message)s
            format = '[%(asctime)s] [%(levelname)s] [%(module)s] [%(lineno)d] %(message)s'

        log = logging.getLogger(filename)
        format_str = logging.Formatter(format, datefmt)
        # backupCount 保存日志的数量，过期自动删除
        # when 按什么日期格式切分(这里方便测试使用的秒)

        # file_handler = FileHandler(filepath, level=loglevel)
        # log.addHandler(file_handler)

        th = TimedRotatingFileHandler(filename=filename, when='D', backupCount=5, encoding='utf-8')
        th.setFormatter(format_str)
        th.setLevel(level)
        log.addHandler(th)

        # 为了看的更视觉效果，可以显示在控制台答应
        cmd = logging.StreamHandler()
        cmd.setFormatter(format_str)
        cmd.setLevel(level)
        log.addHandler(cmd)

        log.setLevel(level)
        return log



import setuptools

with open("README.md", "r", encoding='utf-8') as fh:
    long_description = fh.read()

__author__ = 'aaron'
__date__ = '2020/08/17'

setuptools.setup(
    name="strategies-framework",
    version="1.2.1",
    author="Aaron",
    author_email="103514303@qq.com",
    description="StrategiesFramework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/StarsAaron",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    license='MIT',
    install_requires=['requests', 'logbook', 'zmq'],
    include_package_data=True,
    zip_safe=True
)


import errno
import json
from collections import defaultdict
from queue import Empty, Queue
from threading import Thread
import zmq
from framework.log_handler import DefaultLogHandler
from framework.utils_requests import get_page_json


class ActionFun:
    """
    用户自定义函数封装
    """
    typeName = ''  # 函数名
    fun = None  # 执行的函数
    threadId = None  # 执行线程

    def __init__(self):
        self.typeName = ''
        self.fun = None  # 执行的函数
        self.threadId = None  # 执行线程


class StrategiesManager:
    remote_history_url = 'http://{}/gethistory?code={}&startTime={}&endTime={}&ktype={}'
    threads = {}
    finished = False
    action = []
    tempDir = './temp'
    prepareActionQueue = Queue()

    def __init__(self, configPath='./sconfig.conf', tempPath='./temp'):
        """
        初始化

        配置文件格式：
        {
            "remote_history_url": "127.0.0.1:10039",
            "kl_real_time_addr_in": "tcp://127.0.0.1:10019",
            "kl_real_time_addr_out": "tcp://127.0.0.1:10020"
        }
        :param configPath: 配置文件路径，默认./sconfig.conf
        :param tempPath: 数据缓存文件夹路径，默认./temp
        """
        self.logHandler = DefaultLogHandler(filename='strategies.log')
        self.logHandler.info('************** 策略管理器初始化 ***************')
        self.config = self.file2dict(configPath)
        self.url_ip = self.config['remote_history_url']
        self.kl_real_time_addr_in = self.config['kl_real_time_addr_in']
        self.kl_real_time_addr_out = self.config['kl_real_time_addr_out']
        self.tempDir = tempPath

    def __del__(self):
        self.finished = True
        for k in self.threads.keys():
            self.threads[k].join()
        self.threads.clear()

    def file2dict(self, path):
        """
        从文件读取配置信息转字典
        :param path: 配置文件路径
        :return:
        """
        # 读取配置文件
        with open(path) as f:
            return json.load(f)

    def registAction(self, fun):
        """
        注册回调方法
        :param fun:
        :return:
        """
        af = ActionFun()
        af.typeName = fun.__name__
        af.fun = fun
        self.action.append(af)

    def getHistory(self, stock_code, startTime='', endTime='', ktype='') -> dict:
        """
        获取合约历史数据
        :param stock_code:
        :param startTime:
        :param endTime:
        :param ktype:
        :return:
        """
        url = self.remote_history_url.format(self.url_ip, stock_code, startTime, endTime, ktype)
        rsp = get_page_json(url)
        return rsp

    def _doAction(self, stock_code, startTime='', endTime='', ktype=''):
        self.logHandler.info(f'开启任务 ---> \n历史回测 {stock_code}合约 从{startTime}到{endTime} {ktype} 数据')
        start = startTime.replace('-', '')
        end = endTime.replace('-', '')
        #
        import os
        if not os.path.exists(self.tempDir):
            os.makedirs(self.tempDir)
        #
        filePath = f'{self.tempDir}/{stock_code}-{start}-{end}-{ktype}.dat'
        data_list = None
        if os.path.exists(filePath):
            data_list = self.file2dict(filePath)
        #
        if data_list is None:
            url = self.remote_history_url.format(self.url_ip, stock_code, startTime, endTime, ktype)
            rsp = get_page_json(url)
            data_list = rsp['data']
            if len(data_list) == 0:
                self.logHandler.error(f'从网络获取历史数据返回为空,历史回测 {stock_code}合约 从{startTime}到{endTime} {ktype} 数据')
                return
            # 缓存到本地
            with open(filePath, 'w') as file:
                bJson = json.dumps(data_list, ensure_ascii=False)  # dict转json
                file.writelines(bJson)

        for item in data_list:
            for act in self.action:
                act(item)
        self.logHandler.info('完成任务 --->')
        key = stock_code + startTime + endTime + ktype
        self.threads.pop(key)

    def select_pan_type(self, code: str) -> str:
        """
        判断内外盘
        :param code:NYMEX_F_CL_2007
        :return:
        """
        contratelist = str(code).split('_')
        if contratelist[0].upper() in ['SHFE', 'CFFEX', 'DCE', 'CZCE']:
            return 'in'  # 内盘
        else:
            return 'out'  # 外盘

    def _connect_zmq(self, stock_code, ktype):
        try:
            quote_ctx = zmq.Context()
            client = quote_ctx.socket(zmq.SUB)
            # client.setsockopt(zmq.ZMQ_RECONNECT_IVL, 500)
            # client.setsockopt(zmq.ZMQ_RECONNECT_IVL_MAX, 5000)

            pan_type = self.select_pan_type(stock_code)
            # 请求历史数据
            if pan_type == 'in':
                client.connect(self.kl_real_time_addr_in)
            elif pan_type == 'out':
                client.connect(self.kl_real_time_addr_out)
            else:
                self.logHandler.info("无法判断内外盘")
                return None

            client.setsockopt_string(zmq.SUBSCRIBE, ktype)
            client.setsockopt(zmq.RCVTIMEO, 10000)
            # client.setsockopt_string(zmq.SUBSCRIBE, '3Min')
            # client.setsockopt_string(zmq.SUBSCRIBE, '5Min')
            # client.setsockopt_string(zmq.SUBSCRIBE, '15Min')
            # client.setsockopt_string(zmq.SUBSCRIBE, '30Min')
            # client.setsockopt_string(zmq.SUBSCRIBE, '60Min')
            # client.setsockopt_string(zmq.SUBSCRIBE, 'klDay')
            # client.setsockopt_string(zmq.SUBSCRIBE, 'qt')
            # client.setsockopt_string(zmq.SUBSCRIBE, 'ba')
            # client.setsockopt_string(zmq.SUBSCRIBE, 'ticket')

            return client
        except zmq.error.ZMQError as e:
            self.logHandler.error("zmq 连接出错:%s" % e)
            return None

    def _doRealTimeAction(self, stock_code, ktype):
        self.logHandler.info(f'开启任务 ---> \n实时模拟 {stock_code}合约 {ktype} 数据')
        client = self._connect_zmq(stock_code, ktype)

        for action in self.action:
            self.prepareActionQueue.put(action)

        while not self.finished:
            try:
                if client is not None:
                    response = client.recv()
                else:
                    self.logHandler.error(f'zmq尝试重连......')
                    client = self._connect_zmq(stock_code, ktype)
                    continue
            except zmq.ZMQError as e:
                self.logHandler.error(f'zmq {e.args}')

                client.close()
                self.logHandler.error(f'zmq尝试重连......')
                client = self._connect_zmq(stock_code, ktype)
                continue

            response = str(response, encoding='GB2312')
            if response != ktype:
                self.logHandler.info("response:     {}".format(response))
                code = str(stock_code).split('_')
                temp = response[:20].split(',')
                if code[0] == temp[0] and code[1] == temp[1] and code[2] == temp[2] and code[3] == temp[3]:
                    size = self.prepareActionQueue.qsize()
                    while size != 0:
                        size -= 1
                        if not self.prepareActionQueue.empty():
                            try:
                                doact = self.prepareActionQueue.get(block=False)
                                if doact is not None:
                                    handle_thread = Thread(target=self._wraperFun, name=doact.typeName,
                                                           args=(doact, response))
                                    doact.threadId = handle_thread
                                    handle_thread.start()
                            except Empty as e:
                                self.logHandler.error(f'队列为空:{e.args}')
                                # act(response)
        client.close()
        self.logHandler.info('完成任务 --->')
        key = stock_code + ktype
        self.threads.pop(key)

    def _wraperFun(self, doactfun, data):
        if isinstance(doactfun, ActionFun):
            try:
                doactfun.fun(data)
            except Exception as e:
                self.logHandler.error(f'执行任务<{doactfun.typeName}>发生异常：{e.args}')
            doactfun.threadId = None
            self.prepareActionQueue.put(doactfun)

    def runRealTime(self, stock_code, ktype='1Min'):
        """
        运行实时数据模拟
        :param stock_code:
        :param ktype:
        :return:
        """
        _watch_thread = Thread(target=self._doRealTimeAction, args=[stock_code, ktype], name="runRealTime")
        key = stock_code + ktype
        if self.threads.get(key) is None:
            self.threads[key] = _watch_thread
            _watch_thread.start()
        else:
            self.logHandler.info('任务已存在 --->')

    def runHistory(self, stock_code, startTime='', endTime='', ktype='1Min'):
        """
        运行历史回测
        :param stock_code:
        :param startTime:
        :param endTime:
        :param ktype:
        :return:
        """
        _watch_thread = Thread(target=self._doAction, args=[stock_code, startTime, endTime, ktype], name="runHistory")
        key = stock_code + startTime + endTime + ktype
        if self.threads.get(key) is None:
            self.threads[key] = _watch_thread
            _watch_thread.start()
        else:
            self.logHandler.info('任务已存在 --->')
